package ru.circumflex.core

import java.lang.reflect.InvocationTargetException
import java.util.regex.Pattern
import util.matching.Regex
import javax.servlet._
import javax.servlet.http.{HttpServletResponse, HttpServletRequest}
import java.io.{FileNotFoundException, File}
import org.apache.commons.io.FilenameUtils._

/**
 * ## Circumflex Filters
 *
 * Provides a base class for Circumflex filter implementations.
 */
abstract class AbstractCircumflexFilter extends Filter {

  /**
   * Place your application initialization code here.
   * Does nothing by default.
   */
  def init(filterConfig: FilterConfig) = {}

  /**
   * Place your application shutdown code here.
   * Does nothing by default.
   */
  def destroy = {}

  /**
   * Determines, if a filter should process the request.
   * The default behavior is controlled by `cx.process_?` parameter:
   *
   *  * `String` or `Regex` or `Pattern`;
   *  * the filter processes request if URI *does not match* specified regex;
   *  * `HttpServletRequest => Boolean` or `() => Boolean` functions,
   *  the filter processes request depending on the result of function invocation.
   * </ul>
   */
  def isProcessed(req: HttpServletRequest): Boolean = Circumflex.get("cx.process_?") match {
    case Some(s: String) => !req.getRequestURI.toLowerCase.matches(s)
    case Some(r: Regex) => !req.getRequestURI.toLowerCase.matches(r.toString)
    case Some(p: Pattern) => !p.matcher(req.getRequestURI.toLowerCase).matches()
    case Some(func: (() => Boolean)) => func.apply()
    case Some(func: (HttpServletRequest => Boolean)) => func.apply(req)
    case _ => true
  }

  /**
   * Instantiates a `CircumflexContext` object, binds it to current request,
   * consults `isProcessed, whether the request should be processed
   * and delegates to high-level equivalent
   * `doFilter(CircumflexContext, FilterChain)`
   * if necessary.
   */
  def doFilter(req: ServletRequest, res: ServletResponse, chain: FilterChain): Unit =
    (req, res) match {
      case (req: HttpServletRequest, res: HttpServletResponse) =>
        // try to serve static first
        if (req.getMethod.equalsIgnoreCase("get") || req.getMethod.equalsIgnoreCase("head")) {
          val resource = new File(Circumflex.publicRoot, separatorsToSystem(req.getRequestURI))
          if (resource.isFile) {
            req.getRequestDispatcher(Circumflex.publicUri + req.getRequestURI).forward(req, res)
            return
          }
        }
        if (isProcessed(req)) {
          // Instantiate a context if it does not yet exist and bind it thread-locally.
          CircumflexContext.init(req, res)
          // chain a call and make sure the context is destroyed afterwards
          try {
            doFilter(CircumflexContext.get, chain)
          } finally {
            CircumflexContext.destroy()
          }
        } else chain.doFilter(req, res)
      case _ =>
    }

  /**
   * Implementing classes should provide their filtering logic for processed requests.
   */
  def doFilter(ctx: CircumflexContext, chain: FilterChain): Unit

}

/**
 * Configures Circumflex-based web application and serves it's requests.
 * Web application should configure it according to JSR-154 (Java Servlet Specification) via
 * `your_webapp_root/WEB-INF/web.xml` (also known as Deployment Descriptor).
 */
class CircumflexFilter extends AbstractCircumflexFilter {

  val routerClass: Class[RequestRouter] = Circumflex.get("cx.router") match {
    case Some(s: String) => Circumflex.loadClass[RequestRouter](s)
    case Some(c: Class[RequestRouter]) => c
    case _ => throw new CircumflexException("Could not initialize Request Router; " +
        "configure 'cx.router' properly.")
  }

  /**
   * Executed when no routes match current request.
   * Default behavior is to send 404 NOT FOUND.
   * You may override it, say, to call `chain.doFilter` to pass request
   * along the chain.
   */
  def onNoMatch(ctx: CircumflexContext, chain: FilterChain) =
    ErrorResponse(404, "The requested resource does not exist.")(ctx.response)

  /**
   * Executed when router throws an exception.
   * Default behavior is to send the `500 Internal Server Error` to client.
   */
  def onRouterError(e: Throwable, ctx: CircumflexContext, chain: FilterChain) = {
    cxLog.error("Router threw an exception, see stack trace for details.", e)
    ErrorResponse(500, e.getMessage)(ctx.response)
  }

  /**
   * Executed when `java.io.FileNotFoundException`
   * is thrown from router.
   * Default behavior is to send the `404 Not Found` to client.
   */
  def onNotFound(e: Throwable, ctx: CircumflexContext, chain: FilterChain) = {
    ErrorResponse(404, e.getMessage)(ctx.response)
  }

  /**
   * Instantiates a router that processes current request.
   */
  def doFilter(ctx: CircumflexContext, chain: FilterChain): Unit = {
    cxLog.debug(ctx.request.toString)
    // Set X-Powered-By header
    ctx.response.setHeader("X-Powered-By", "Circumflex v. 1.2")
    // Set character encoding
    ctx.request.setCharacterEncoding("UTF-8")
    try {
      routerClass.getConstructor().newInstance()
      // Request not matched by router
      cxLog.debug("No routes matched.")
      onNoMatch(ctx, chain)
    } catch {
      case e: InvocationTargetException => e.getCause match {
        case e: RouteMatchedException =>
          // Request matched
          e.response.apply(ctx.response)
          cxLog.debug(ctx.response.toString)
        case e: FileNotFoundException =>
          // Not found file or template
          onNotFound(e, ctx, chain)
          cxLog.debug("File not found: " + e.getMessage)
        case e =>  // Generic error
          onRouterError(e, ctx, chain)
      }
    }
  }

  /**
   * Called when a filter is instantiated by Servlet Container.
   */
  override def init(cfg: FilterConfig) = {
    cxLog.info("Circumflex v. 1.2")
  }

}
